// =======================================================================
//
//   povany_tile.c
//
//   written january 2004 by Christoph Hormann <chris_hormann@gmx.de>
//
//   combines partial renders from POV into a large image
//
// -----------------------------------------------------------------------
//
//   Copyright (C) 2004-2005 by Christoph Hormann <chris_hormann@gmx.de>
//
//   All rights reserved.
//
//   Redistribution and use in source and binary forms, with or without
//   modification, are permitted provided that the following conditions
//   are met:
//
//    * Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    * Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//
//   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
//   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
//   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
//   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
//   CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
//   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
//   ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
//   POSSIBILITY OF SUCH DAMAGE.
//
// =======================================================================

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <getopt.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>

#include "povany.h"

#define PROG_STRING "POV-Anywhere Tile 0.3 alpha\nCopyright 2004-2005 by Christoph Hormann <chris_hormann@gmx.de>"

int Input_Count = 0;

int Debug = FALSE;

char *dimensions_string = NULL;
char *write_filename = NULL;

void RGB_Pos_Input(RGB_Int rgb, TILE_INPUT *Input, int Pos_x, int Pos_y);
TILE_INPUT *create_input(char *fnm, TILE_INPUT *Inputs);

void usage ( void ) {
  fprintf(stderr, "\ncombines partial renders from POV-Ray into a large image \n\n");
  fprintf(stderr, "Usage: \n");
  fprintf(stderr, "  povtile \\\n");
  fprintf(stderr, "    -g tile_width:tile_height[:total_width:total_height] \\\n");
  fprintf(stderr, "    -i input_file ... [-o <output_file>] [OPTIONS]\n"); 
  fprintf(stderr, "\n");
  fprintf(stderr, "  input_file has to be in binary ppm or pgm format,\n"); 
  fprintf(stderr, "  if total_width and total_height are omitted they are\n"); 
  fprintf(stderr, "  determined form the input files\n");
  fprintf(stderr, "  if output_file is not specified output is written to stdout\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "Options:\n");
  fprintf(stderr, "  -d : write additional debug output\n");
  fprintf(stderr, "  -v : only print version and exit\n");
  fprintf(stderr, "  -h : print this usage info and exit\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "DISCLAIMER:\n");
  fprintf(stderr, "  This program comes without any warranty, especially\n");
  fprintf(stderr, "  do not rely on it for security sensitive applications.\n");
  fprintf(stderr, "\n");
}

void Error (TILE_INPUT *Inputs, char *format,...) {

  va_list marker;
  char localvsbuffer[1024];
  int i;

  va_start(marker, format);
  vsprintf(localvsbuffer, format, marker);
  va_end(marker);

  fprintf(stderr, "Error: %s\n", localvsbuffer);
  if (Input_Count > 0)
  {
    for (i=0;i<Input_Count;i++)
    {
      if ( Inputs[i].file_handle ) fclose(Inputs[i].file_handle);
    }
    free(Inputs);
  }

  exit(1);
}


int main(int argc,char **argv)
{
  TILE_INPUT *Inputs = NULL;

  char *Pos;
  char buffer[1024];
  int c, i;
  int tile_width = 0; 
  int tile_height = 0; 
  int total_width = 0;
  int total_height = 0;
  int total_rgb = FALSE;
  int total_depth = 0;
  RGB_Int rgb_vals;
  RGB_Byte rgb_vals8;

  int y_cnt, x_cnt;
  int iposy, iposx;
  int ix_idx, iy_idx, ix_cnt, iy_cnt;
  int Idx;
  int Tile_Count_X, Tile_Count_Y;
  FILE *fout;
  int fout_open = FALSE;


  fprintf(stderr, "%s\n", PROG_STRING);

  while ( (c = getopt(argc,argv,"d?vhi:g:o:")) != -1 )
  {
    switch (c)
    {
      case 'v' : exit(0);   break;
      case 'g' : dimensions_string = strdup(optarg);  break;
      case 'i' : Inputs = create_input(strdup(optarg), Inputs);   break;
      case 'o' : write_filename   = strdup(optarg);   break;
      case 'd' : Debug = TRUE;  break;
      case 'h' :
      case '?' :
      default  :
        usage();
        return 0;
    }
  }

  if (Input_Count==0) 
  {
    usage();
    fprintf(stderr, "  no input files specified\n\n");
    return 1;
  }

  // prepare dimensions_string if given
  if (dimensions_string != NULL)
  {
    strncpy(buffer, dimensions_string, strlen(dimensions_string)>1024?1024:strlen(dimensions_string));
    Pos = strtok(buffer, ":");
    if (Pos != NULL) tile_width = atoi(Pos);
    Pos = strtok(NULL, ":");
    if (Pos != NULL) tile_height = atoi(Pos);
    Pos = strtok(NULL, ":");
    if (Pos != NULL) total_width = atoi(Pos);
    Pos = strtok(NULL, ":");
    if (Pos != NULL) total_height = atoi(Pos);

    if (Inputs[0].width )
    for (i=0;i<Input_Count;i++)
    {
      if ((Inputs[i].width < tile_width) || (Inputs[i].height < tile_height))
        Error(Inputs, "  image file %s is smaller than the given tile size\n", Inputs[i].file_name);
      if ((Inputs[i].width < total_width) || (Inputs[i].height < total_height))
        Inputs[i].partial = FALSE;
      else Inputs[i].partial = TRUE;
      total_rgb = total_rgb || Inputs[i].rgb;
      total_depth = total_depth>Inputs[i].depth?total_depth:Inputs[i].depth;
      //if (Inputs[i].rgb != Inputs[0].rgb)
      //  Error(Inputs, "  Error: all input images must be same subformat (%s)\n", Inputs[i].file_name);
    }

    if (total_width==0) total_width = Inputs[0].width;
    if (total_height==0) total_height = Inputs[0].height;

  }
  else Error(Inputs, "  Error: no tiling geometry given\n");

  Tile_Count_Y=(int)ceil((double)total_height/(double)tile_height);
  Tile_Count_X=(int)ceil((double)total_width/(double)tile_width);

  if (Input_Count < Tile_Count_X*Tile_Count_Y)
    Error(Inputs, "  Error:  %d input tiles required but only %d given\n", Tile_Count_X*Tile_Count_Y, Input_Count);

  if (Debug)
  {
    fprintf(stderr, "\nrunning with settings:\n");
    for (i=0;i<Input_Count;i++)
    {
      fprintf(stderr, " input file <%s>:\n", Inputs[i].file_name);
      fprintf(stderr, "     - width:%d\n", Inputs[i].width);
      fprintf(stderr, "     - height:%d\n", Inputs[i].height);
      fprintf(stderr, "     - colors:%d\n", Inputs[i].depth);
      fprintf(stderr, "     - rgb:%d\n", Inputs[i].rgb);
      fprintf(stderr, "     - partial:%d\n", Inputs[i].partial);
    }

    if (write_filename != NULL)
      fprintf(stderr, "\n output file <%s>:\n", write_filename);
    else 
      fprintf(stderr, "\n output file <stdout>:\n");
    fprintf(stderr, "     - width:%d\n", total_width);
    fprintf(stderr, "     - height:%d\n", total_height);
    fprintf(stderr, "     - colors:%d\n", total_depth);
    fprintf(stderr, "     - rgb:%d\n", total_rgb);
    fprintf(stderr, "\n");
  }

  fprintf(stderr, "\n");

  if (write_filename != NULL)
  {
    fout = fopen(write_filename,"w");

    if ( ! fout ) Error(Inputs, "  can't open file %s for writing\n", write_filename);

    fout_open = TRUE;

  } else fout = stdout;

  fprintf(stderr, "Combining %d tiles into one image...\n", Input_Count);

  if (total_rgb) fprintf(fout,"P6\n");
  else fprintf(fout,"P5\n");
  fprintf(fout,"%d %d\n",total_width, total_height);
  fprintf(fout,"%d\n", total_depth);

  for (y_cnt=0;y_cnt<total_height;y_cnt++)
  {
    iy_idx = y_cnt / tile_height;
    iy_cnt = y_cnt % tile_height;

    for (x_cnt=0;x_cnt<total_width;x_cnt++)
    {
      ix_idx = x_cnt / tile_width;
      ix_cnt = x_cnt % tile_width;

      Idx = ix_idx + iy_idx*Tile_Count_X;

      if (Inputs[Idx].partial)
      {
        iposx = x_cnt;
        iposy = iy_cnt;
      }
      else
      {
        iposx = ix_cnt;
        iposy = iy_cnt;
      }

      RGB_Pos_Input(rgb_vals, &Inputs[Idx], iposx, iposy);

      if (total_rgb)
      {
        if (total_depth>255)
        {
          if (fwrite(&rgb_vals,6,1,fout) != 1)
            fprintf(stderr, "Warning: error writing output\n");
        }
        else
        {
          rgb_vals8[0] = rgb_vals[0];
          rgb_vals8[1] = rgb_vals[1];
          rgb_vals8[2] = rgb_vals[2];
          if (fwrite(&rgb_vals8,3,1,fout) != 1)
            fprintf(stderr, "Warning: error writing output\n");
        }
      }
      else
      {
        if (total_depth>255)
        {
          if (fwrite(&rgb_vals,2,1,fout) != 1)
            fprintf(stderr, "Warning: error writing output\n");
        }
        else
        {
          rgb_vals8[0] = rgb_vals[0];
          if (fwrite(&rgb_vals8,1,1,fout) != 1)
            fprintf(stderr, "Warning: error writing output\n");
        }
      }

    }

    if (total_height>100)
      if (y_cnt % (total_height/100) < 2) 
      {
        fprintf(stderr, "\r%d percent", (int)( ((double)y_cnt / (double)total_height)*100 ));
      }
  }

  // clean up
  if (fout_open) fclose(fout);
  for (i=0;i<Input_Count;i++)
  {
    if ( Inputs[i].file_handle ) fclose(Inputs[i].file_handle);
  }

  free(Inputs);

  fprintf(stderr, "\nFinished!\n");

  return 0;
}

void RGB_Pos_Input(RGB_Int rgb, TILE_INPUT *Input, int Pos_x, int Pos_y)
{
  RGB_8 Bf_rgb8;
  RGB_16 Bf_rgb16;
  unsigned char Bf_char;
  unsigned char Bf_chars[2];

  if (Input->rgb)
  {
    if (Input->depth<256)  // 3x8bit rgb
    {
      fseek(Input->file_handle, Input->start_pos+(Input->width*Pos_y+Pos_x)*3, 0);
      if (fread(&Bf_rgb8, 1, 3, Input->file_handle)<3)
      {
        rgb[0] = 0;
        rgb[1] = 0;
        rgb[2] = 0;
      }
      else
      {
        rgb[0] = Bf_rgb8[0];
        rgb[1] = Bf_rgb8[1];
        rgb[2] = Bf_rgb8[2];
      }
    }
    else                // 3x16bit rgb
    {
      fseek(Input->file_handle, Input->start_pos+(Input->width*Pos_y+Pos_x)*6, 0);
      if (fread(Bf_rgb16, 1, 6, Input->file_handle)<6)
      {
        rgb[0] = 0;
        rgb[1] = 0;
        rgb[2] = 0;
      }
      else
      {
        rgb[0] = Bf_rgb16[0][0] + Bf_rgb16[0][1]*256;
        rgb[1] = Bf_rgb16[1][0] + Bf_rgb16[1][1]*256;
        rgb[2] = Bf_rgb16[2][0] + Bf_rgb16[2][1]*256;
      }
    }
  }
  else
  {
    if (Input->depth<256)  // 1x8bit gray
    {
      fseek(Input->file_handle, Input->start_pos+(Input->width*Pos_y+Pos_x), 0);
      if (fread(&Bf_char, 1, 1, Input->file_handle)<1)
      {
        rgb[0] = 0;
        rgb[1] = 0;
        rgb[2] = 0;
      }
      else
      {
        rgb[0] = Bf_char;
        rgb[1] = rgb[0];
        rgb[2] = rgb[0];
      }
    }
    else                // 1x16bit gray
    {
      fseek(Input->file_handle, Input->start_pos+(Input->width*Pos_y+Pos_x)*2, 0);
      if (fread(Bf_chars, 1, 2, Input->file_handle)<2)
      {
        rgb[0] = 0;
        rgb[1] = 0;
        rgb[2] = 0;
      }
      else
      {
        rgb[0] = Bf_chars[0] + Bf_chars[1]*256;
        rgb[1] = rgb[0];
        rgb[2] = rgb[0];
      }
    }
  }

  return;
}

TILE_INPUT *create_input(char *fnm, TILE_INPUT *Inputs)
{
  unsigned char header[2];
  char line[1024];
  char *ptr;
  char *Pos;

  int Idx=Input_Count;

  // allocate memory
  Input_Count++;
  Inputs = realloc(Inputs, Input_Count*sizeof(TILE_INPUT));

  strcpy(Inputs[Idx].file_name, fnm);

  Inputs[Idx].file_handle = fopen(Inputs[Idx].file_name,"r");

  if ( ! Inputs[Idx].file_handle )
    Error(Inputs, "  Cannot open file %s.\n", Inputs[Idx].file_name);

  // --- Read Header ---
  if (fread(header, 1, 2, Inputs[Idx].file_handle)<2)
    Error(Inputs, "  Cannot read header of PPM image %s.", Inputs[Idx].file_name);

  if(header[0] != 'P') Error(Inputs, "  File %s is not in PPM format.", Inputs[Idx].file_name);

  do
  {
    if (fgets(line, 1024, Inputs[Idx].file_handle) == NULL)
      Error(Inputs, "  Error reading from file %s.\n", Inputs[Idx].file_name);
    Pos = strchr(line, '\n');           // remove line break
    if (Pos != NULL) Pos[0] = '\0';
    else line[1023] = '\0';
    if ((ptr = strchr(line, '#')) != NULL) *ptr = '\0';  // remove comment 
  }
  while (line[0]=='\0');  // read until line without comment from beginning 

  // --- First: two numbers: with and height --- 
  if (sscanf(line,"%d %d",&Inputs[Idx].width, &Inputs[Idx].height) != 2)
    Error(Inputs, "  Cannot read width and height from PPM image %s.", Inputs[Idx].file_name);

  if (Inputs[Idx].width <= 0 || Inputs[Idx].height <= 0)
    Error(Inputs, "  Invalid width or height read from PPM image %s.", Inputs[Idx].file_name);

  do
  {
    if (fgets(line, 1024, Inputs[Idx].file_handle) == NULL)
      Error(Inputs, "  Error reading from file %s.\n", Inputs[Idx].file_name);
    Pos = strchr(line, '\n');           // remove line break
    if (Pos != NULL) Pos[0] = '\0';
    else line[1023] = '\0';
    if ((ptr = strchr(line, '#')) != NULL) *ptr = '\0';  // remove comment 
  }
  while (line[0]=='\0');  // read until line without comment from beginning 

  // --- Second: one number: color depth --- 
  if (sscanf(line,"%d",&Inputs[Idx].depth) != 1)
    Error(Inputs, "  Cannot read color depth from PPM image.");

  if ((Inputs[Idx].depth > 65535) || (Inputs[Idx].depth < 1))
    Error(Inputs, "  Unsupported number of colors (%d) in PPM image.", Inputs[Idx].depth);

  Inputs[Idx].start_pos = ftell(Inputs[Idx].file_handle);

  switch (header[1])
  {
    /*
    case '2':
      Inputs[Idx].rgb = FALSE;
      break;
    case '3':
      Inputs[Idx].rgb = TRUE;
      break;
    */
    case '5':
      Inputs[Idx].rgb = FALSE;
      break;
    case '6':
      Inputs[Idx].rgb = TRUE;
      break;
    default:
      Error(Inputs, "  File %s is not in PPM format (type %d).", Inputs[Idx].file_name, header[1]);
      break;
  }

  return Inputs;

}


